#include "../../internal-include/CD/cd_internal.hpp"
#include <PSX/System/IOPorts/dma_io.hpp>
#include <PSX/System/IOPorts/interrupt_io.hpp>
#include <PSX/System/syscalls.hpp>
#include <stdio.hpp>
struct Registers {
    union {
        struct {
            uint32_t r0, at, v0, v1, a0, a1, a2, a3;
            uint32_t t0, t1, t2, t3, t4, t5, t6, t7;
            uint32_t s0, s1, s2, s3, s4, s5, s6, s7;
            uint32_t t8, t9, k0, k1, gp, sp, fp, ra;
        } n;
        uint32_t r[32];
    } GPR;
    uint32_t returnPC;
    uint32_t hi, lo;
    uint32_t SR;
    uint32_t Cause;
};

struct Thread {
    uint32_t flags, flags2;
    struct Registers registers;
    uint32_t unknown[9];
};

struct Process {
    struct Thread* thread;
};

struct Globals {
    /* 100 */ struct HandlersStorage* handlersArray;
    /* 104 */ uint32_t handlersArraySize;
    /* 108 */ struct Process* processes;
    /* 10c */ uint32_t processBlockSize;
    /* 110 */ struct Thread* threads;
    /* 114 */ uint32_t threadBlockSize;
    /* 118 */ uint32_t xxx_06;
    /* 11c */ uint32_t xxx_07;
    /* 120 */ struct EventInfo* events;
    /* 124 */ uint32_t eventsSize;
    /* 128 */ uint32_t xxx_0a;
    /* 12c */ uint32_t xxx_0b;
    /* 130 */ uint32_t xxx_0c;
    /* 134 */ uint32_t xxx_0d;
    /* 138 */ uint32_t xxx_0e;
    /* 13c */ uint32_t xxx_0f;
    /* 140 */ struct File* files;
    /* 144 */ uint32_t filesSize;
    /* 148 */ uint32_t xxx_12;
    /* 14c */ uint32_t xxx_13;
    /* 150 */ struct Device* devices;
    /* 154 */ struct Device* devicesEnd;
    /* 158 */ uint32_t xxx_16;
    /* 15c */ uint32_t xxx_17;
    /* 160 */ uint32_t xxx_18;
    /* 164 */ uint32_t xxx_19;
    /* 168 */ uint32_t xxx_1a;
    /* 16c */ uint32_t xxx_1b;
    /* 170 */ uint32_t xxx_1c;
    /* 174 */ uint32_t xxx_1d;
    /* 178 */ uint32_t xxx_1e;
    /* 17c */ uint32_t xxx_1f;
};

Globals *__globals;

namespace JabyEngine {
            struct ToT {
	unsigned long *head;
	long size;
};

struct TCBH {
	TCB *entry;	/* NULL */
	long flag;
};

TCBH *exec_TCB;
TCB  *mainFlow;    

inline unsigned short getSystemTableNumber(uint32_t thread)
{
	return (thread & 0xFFFF);
}

inline TCB* getTCB(uint32_t thread)
{
	return (TCB*) (((ToT*)0x100)[2]).head + getSystemTableNumber(thread);
}

int changeThread(int threadId) { return SysCall::changeThreadSubFunction((uint32_t)&__globals->threads[threadId & 0xffff]); }

    namespace CD {
        namespace internal {
            static constexpr auto AudioSectorMode = CD_IO::Mode::from(CD_IO::Mode::SingleSpeed, CD_IO::Mode::AutoPauseTrack, CD_IO::Mode::CDDA);
            static constexpr auto DataSectorMode  = CD_IO::Mode::from(CD_IO::Mode::DoubleSpeed, CD_IO::Mode::DataSector);
            
            static SysCall::InterruptVerifierResult interrupt_verifier();
            static uint32_t interrupt_handler(uint32_t);

            static uint32_t stack[1024];
            static uint32_t thread = 0;

            static SectorBufferAllocator sector_allocator;
            static uint32_t              cur_lba;
            static uint32_t              dst_lba;
            
            uint8_t                   cmd_interrupt_bit = 0;
            State                     current_state     = State::Free;
            SysCall::InterrupCallback callback          = {
                .next              = nullptr,
                .handler_function  = interrupt_handler,
                .verifier_function = interrupt_verifier
            };

            static void pause_cd() {
                CD_IO::PortIndex0::change_to();
                Command::send<CD_IO::PortIndex0>(CD_IO::Command::Pause);
            }

            // Requires Index0
            static void read_cd(uint32_t lba) {
                const auto loc = CDTimeStamp::from(lba);               

                Command::send_wait<CD_IO::PortIndex0>(CD_IO::Command::SetLoc, loc.get_min_cd(), loc.get_sec_cd(), loc.get_sector_cd());
                Command::send<CD_IO::PortIndex0>(CD_IO::Command::ReadN);
                current_state = State::Reading;
            }

            static void read_sector_dma(CD_IO::DataSector& sector) {
                static const auto WaitSectorReady = []() {
                    while(!CD_IO::IndexStatus.read().is_set(CD_IO::IndexStatus::HasDataFifoData));
                };

                static const auto ReadSector = [](uint32_t* dst) {
                    DMA_IO::CDROM.set_adr(reinterpret_cast<uintptr_t>(dst));
                    DMA_IO::CDROM.block_ctrl.write(DMA_IO::BCR::SyncMode0::for_cd());
                    DMA_IO::CDROM.channel_ctrl.write(DMA_IO::CHCHR::StartCDROM());
                    
                    DMA_IO::CDROM.wait();

                    CD_IO::PortIndex0::Request.write(CD_IO::Request::reset());
                };

                WaitSectorReady();
                ReadSector(sector.data);
            }

            static void read_sector_to(CD_IO::DataSector& sector) {
                CD_IO::PortIndex0::change_to();
                CD_IO::PortIndex0::Request.write(CD_IO::Request::want_data());

                // We only support DMA rn
                read_sector_dma(sector);

                // Do we ever want to support reading via IO Port?
                // Doesn't seem to important when we can use DMA
            }

            static SysCall::InterruptVerifierResult interrupt_verifier() {
                if(Interrupt::is_irq(Interrupt::CDROM)) {
                    return SysCall::InterruptVerifierResult::ExecuteHandler;
                }
                
                else {
                    return SysCall::InterruptVerifierResult::SkipHandler;
                }
            }

            static uint32_t interrupt_handler(uint32_t) {
                const auto old_status = CD_IO::IndexStatus.read();

                CD_IO::PortIndex1::change_to();
                const auto cur_irq = CD_IO::Interrupt::get_type(CD_IO::PortIndex1::InterruptFlag);
                CD_IO::Interrupt::ack_extended(CD_IO::PortIndex1::InterruptFlag);

                cmd_interrupt_bit = bit::clear(cmd_interrupt_bit, cur_irq);
                switch(cur_irq) {
                    case CD_IO::Interrupt::DataReady: {
                        // Obtain sector content here
                        auto* sector = sector_allocator.allocate_sector();
                        if(sector) {
                            //Now obtain sector
                            read_sector_to(*sector);

                            cur_lba++;
                            if(cur_lba == dst_lba) {
                                current_state = State::Done;
                                pause_cd();
                            }
                        }

                        else {
                            current_state = State::BufferFull;
                            pause_cd();
                        }
                    } break;

                    case CD_IO::Interrupt::DataEnd: {
                        
                        //set_tcb(get_tcb_of_th(thread));

                        /*printf("EPC: 0x%p; SR: 0x%p\n", get_current_tcb().reg[31], get_current_tcb().sr);
                        get_current_tcb().reg[31] = (uint32_t)(void(*)())[]() {
                            printf("Wuff! Now crash!\n");
                            while(true);
                        };
                        printf("EPC: 0x%p; SR: 0x%p\n", get_current_tcb().reg[31], get_current_tcb().sr); while(true);*/
                        //SysCall::ReturnFromException();

                        CD_IO::PortIndex0::change_to();
                        Command::send<CD_IO::PortIndex0>(CD_IO::Command::SetLoc, static_cast<uint8_t>(0x0), static_cast<uint8_t>(0x09), static_cast<uint8_t>(0x0));

                        CD_IO::PortIndex1::change_to();
                        while(CD_IO::Interrupt::get_type(CD_IO::PortIndex1::InterruptFlag) != CD_IO::Interrupt::Acknowledge);
                        CD_IO::Interrupt::ack_extended(CD_IO::PortIndex1::InterruptFlag);

                        CD_IO::PortIndex0::change_to();
                        Command::send<CD_IO::PortIndex0>(CD_IO::Command::Play);
                    } break;

                    case CD_IO::Interrupt::DiskError: {
                        current_state = State::Error;
                    } break;
                }

                // No masking required because we can only write bit 0 - 2
                CD_IO::IndexStatus.write(old_status);
                Interrupt::ack_irq(Interrupt::CDROM);

                if(thread != 0) {
                    //changeThread(thread);
                    exec_TCB->entry = reinterpret_cast<TCB*>(&__globals->threads[thread & 0xffff]);
                }
                SysCall::ReturnFromException();
                __builtin_unreachable();
            }

            void read_file(AutoLBAEntry file_info, const SectorBufferAllocator& buffer_allocator) {
                cur_lba          = file_info.get_lba();
                dst_lba          = cur_lba + file_info.get_size_in_sectors();
                sector_allocator = buffer_allocator;

                Command::wait_completed();
                CD_IO::PortIndex0::change_to();
                Command::send_wait<CD_IO::PortIndex0>(CD_IO::Command::SetMode, DataSectorMode);

                read_cd(cur_lba);
            }

            void continue_reading() {
                if(current_state == State::BufferFull) {
                    Command::wait_completed();
                    read_cd(cur_lba);    
                }                
            }



            void enable_CDDA() {
                if(thread == 0) {
                    __globals = reinterpret_cast<Globals*>(0x100);
                    ToT *table = (ToT*)0x100;
                    exec_TCB = (TCBH*)table[1].head;
                    mainFlow = exec_TCB->entry;

                    SysCall::EnterCriticalSection();
                    thread = SysCall::OpenTh([]() {
                        printf("Hallo!\n");
                        printf("Me: 0x%p - 0x%p\n", get_tcb_of_th(thread), &get_current_tcb());

                        exec_TCB->entry = mainFlow;
                        SysCall::ReturnFromException();
                        //changeThread(0);
                    }, reinterpret_cast<uint32_t*>(0x80010000), reinterpret_cast<uint32_t*>(mainFlow->reg[28]));
                    SysCall::ExitCriticalSection();

                    printf("Raw: %x, TCB: %i\n", thread, thread & 0xFFFF);
                    //__globals->threads[thread & 0xffff].registers.SR = 0x404;

                    //mainFlow = exec_TCB->entry;
                    //changeThread(thread);
                }

                Command::wait_completed();
                CD_IO::PortIndex0::change_to();
                Command::send_wait<CD_IO::PortIndex0>(CD_IO::Command::SetMode, AudioSectorMode);
            }
        }
    }
}